from PIL import Image
from .util import handle_images
from .image import image_router
from .text import text_router


def image_in_image(*images, key: int=None, ratio: int=4):
    """Hides image(s) in an image or extracts hidden image(s) from an image.
    
    If multiple images passed:
        Hides every image except the first one, in the first image and returns it.
    If only an image is passed:
        Extracts images from the given image and returns them.
        
    Parameters
    ----------
    images : list[np.ndarray]
        A list of images represented with numpy arrays.
    key : int = None
        A key that will be used for RNG.
    ratio : int = 4
        Tells how many bits will be from the cover image. 
        Higher ratio is higher quality cover image and lesser 
        quality hidden images. 4 is default and it means the same quality
        for cover and hidden images.
    """
    if len(images) == 0:
        raise Exception("Error: No image passed.")
    images = handle_images(images)
    return image_router(*images, key=key, ratio=ratio)
    
    
def text_in_image(image: Image.Image, 
                  key: int=None, 
                  text: str=None, 
                  nnoise: int=1):
    """Hides text in an image or extracts hidden text from an image.
    
    If text is passed:
        Encodes the text inside the image using the key.
        The same key will be used to decode the data later.
    If no text:
        Extracts a text from the image using the key.
        
    Parameters
    ----------
    image : numpy.ndarray
        An image represented in a numpy array.
    key : int = None
        A key that will be used for RNG.
    text : str = None
        The piece of text that will be encoded.
    nnoise : int
        Used when encoding an image with text. 
        This is a coefficent to determine the size of the salt.
        salt size = nnoise * size of the encoded data.
    """
    image = handle_images([image])[0]
    if not key:
        raise Exception("Error: No key is passed.")
    return text_router(image=image, key=key, text=text, nnoise=nnoise)



import numpy as np
from .hp import IMAGE_SIZE, BYTE
from PIL import Image
from typing import List


def image_router(*images, ratio: int, key: int=None):
    """Just a simple encapsulation, logic is described in the core file."""
    if ratio <= 0 or ratio >= BYTE:
        raise Exception("Error: Excepted values for ratio are in the range [1,8)")
    if key:
        if len(images) == 1:
            return _decode_image_with_key(images[0], key=key, ratio=ratio)
        return _encode_images_with_key(images, key=key, ratio=ratio)
    # temp
    raise Exception("Error: No key methods are not useful and will be removed.")
    if len(images) == 1:
        return _decode_image_no_key(images[0])
    return _encode_images_no_key(images)
    

def _encode_images_with_key(org_images: List[np.ndarray], 
                            key: int, 
                            ratio: int) -> Image.Image:
    """Takes a list of images and combines them. Every image 
    except the first one is hid into the first image.

    The hiding is done by storing the 4 most significant bits of 
    the hidden images in the randomly chosen places of first 
    image's least significant bits.
    
    Parameters
    ----------
    images : list
        A list of images to be combined. First is the "cover", the
        rest is hidden.
    key : int
        The key is needed for decoding the image later.
    ratio : int
        Tells how many bits are from the cover image.

    Returns
    -------
    PIL.Image.Image
        A cover image with hidden images in it.
    """
    # the function works with the flattened data
    images = [im.reshape(np.prod(im.shape)).copy() for im in org_images]
    # the cover image
    data = images[0]
    # generating the indexes using key
    np.random.seed(key)
    final_idxs = []
    full_idxs = []
    for _ in range(len(images)-1):
        availables = np.setdiff1d(np.arange(data.shape[0]), full_idxs)
        idxs = np.random.choice(availables, np.prod((*IMAGE_SIZE, 3)), replace=False)
        full_idxs = [*full_idxs, *idxs]
        final_idxs.append(idxs)
    # making the magic happen using bit shifting operators
    for i in range(len(images)-1):
        data[final_idxs[i]] = (data[final_idxs[i]] >> (BYTE-ratio) << (BYTE-ratio)) + \
                              (images[i+1] >> ratio)
    return Image.fromarray(data.reshape(org_images[0].shape))


def _decode_image_with_key(image: np.ndarray, 
                           key: int, 
                           ratio: int) -> List[Image.Image]:
    """Takes an image with hidden images in it and returns 
    a list of hidden images.
    
    Parameters
    ----------
    images : list
        A list of images to be combined. First is the "cover", the
        rest is hidden.
    key : int
        The key used when encoding the data.
    ratio : int
        Tells how many bits are from the cover image.

    Returns
    -------
    list[PIL.Image.Image]
        List of extracted images.
    """
    # determining the max number of hiddens
    num_hidden = np.prod(image.shape) // np.prod((*IMAGE_SIZE, 3))
    # the function works with the flattened data
    data = image.reshape(np.prod(image.shape)).copy()
    # generating the indexes using key
    np.random.seed(key)
    final_idxs = []
    full_idxs = []
    for _ in range(num_hidden):
        availables = np.setdiff1d(np.arange(data.shape[0]), full_idxs)
        idxs = np.random.choice(availables, np.prod((*IMAGE_SIZE, 3)), replace=False)
        full_idxs = [*full_idxs, *idxs]
        final_idxs.append(idxs)
    # getting hidden images from the right indexes using bitshifting
    images = []
    for i in range(num_hidden):
        images.append(Image.fromarray((data[final_idxs[i]] << ratio).reshape((*IMAGE_SIZE, 3))))
    return images


def _encode_images_no_key(images: List[np.ndarray]) -> np.ndarray:
    """Takes a list of images and combines them. Every image 
    is combined into an image that looks like the first image.

    The combination is done by combining as much significant 
        bits as possible per image.

    If two images -> combines their 4 most significant digits.
    If three images -> combines 3 most from the first two images
        and 2 most from the last image.
    If four images -> combines their 2 most significant digits.

    Parameters
    ----------
    images : list
        A list of images to be combined.

    Returns
    -------
    list[numpy.ndarray]
        a list of images in the form of numpy arrays.
    """
    if len(images) == 4:
        return sum([image >> 6 << i for (image, i) in zip(images, range(6, -1, -2))])
    if len(images) == 3:
        return (images[0] >> 5 << 5) + (images[1] >> 5 << 2) + (images[2] >> 6)
    if len(images) == 2:
        return (images[0] >> 4 << 4) + (images[1] >> 4)
    raise Exception("Error: Number of images more than 4 is not supported.")


def _decode_image_no_key(combined_image: np.ndarray,
                         num_hidden: int) -> List[Image.Image]:
    """Extracts and returns the `num_hidden` images from the given image.

    Parameters
    ----------
    combined_image : numpy.ndarray
        An image represented by a numpy array.
    num_hidden : int
        Total number of images to be extracted.

    Returns
    -------
    list[PIL.Image.Image]
        a list of extracted images.
    """
    if num_hidden == 4:
        return [Image.fromarray(combined_image >> i << 6) for i in range(6, -1, -2)]
    if num_hidden == 3:
        return [Image.fromarray(im) for im in [(combined_image >> 5 << 5),
                                               (combined_image >> 2 << 5),
                                               (combined_image << 6)]]
    if num_hidden == 2:
        return [Image.fromarray(combined_image >> 4 << 4), Image.fromarray(combined_image << 4)]
    raise Exception("Range of the number of images is [2, 4].")


import requests
import numpy as np
from PIL import Image
from typing import List
from .hp import IMAGE_SIZE


def _from_url(url: str) -> Image.Image:
    """Takes an image url, scrapes the image, and returns it as a numpy array."""
    try:
        return Image.open(requests.get(url, stream=True).raw)
    except:
        raise Exception("Error: Please provide a valid image URL.")


def handle_images(images: list) -> List[np.ndarray]:
    """Given a list of elements(images or image URLs) or both mixed, 
    converts them to numpy arrays and returns. All images are resized 
    except the original(the first) one.

    Parameters
    ----------
    images : list
        A list of mixed types. Supported types:
            -PIL.Image.Image
            -numpy.ndarray
            -str (an image url)

    Returns
    -------
    list[numpy.ndarray]
        a list of images in the form of numpy arrays.
    """
    fixed = []
    mapping = {Image.Image: lambda x: x,
               np.ndarray: Image.fromarray, str: _from_url}
    for im in images:
        if type(im) in mapping.keys():
            fixed.append(mapping[type(im)](im))
        elif Image.isImageType(im):
            fixed.append(Image.fromarray(np.array(im)))
        else:
            print(type(im))
            raise TypeError("Error: Unknown image type.")
    # resizing the hidden images to the predetermined size
    fixed[1:] = [im.resize(IMAGE_SIZE) for im in fixed[1:]]
    fixed = [np.array(im) for im in fixed]
    # now it will check for the volumes if images will be combined
    if len(fixed) > 1:
        volumes = [np.prod(im.shape) for im in fixed]
        if volumes[0] < sum(volumes[1:]):
            raise Exception("""Error: Not enough space to put images. 
                Either decrease the number of images or decrease the 
                hyperparameter `IMAGE_SIZE`""")
    return fixed



import numpy as np
import warnings
from PIL import Image
from .hp import MAX_TEXT_LENGTH, BYTE

def text_router(image: Image.Image, 
                key: int=None, 
                text: str=None, 
                nnoise: int=1):
    """Just a simple encapsulation, logic is described in the core file."""
    if text:
        if len(text) > MAX_TEXT_LENGTH:
            raise Exception("Error: Text is too long.")
        return _encode_image_with_text(image, text=text, key=key, nnoise=nnoise)
    return _decode_image(image, key=key)


def _text_to_binary(text: str) -> np.ndarray:
    """Takes a string and converts it to a list of 1's and 0's.
    Any character with ASCII code bigger than 2^8 will be neglected.
    """
    # chars to ascii
    ord_text = [ord(ch) for ch in text]
    if max(ord_text) > 255:
        warnings.warn("""
            Warning: Chars with ascii codes higher than 
                     255 is not supported and will be lost""")
        ord_text = [min(ord_, 255) for ord_ in ord_text]
    # padding the text with whitespace until it reaches the max length
    ord_text = ord_text + [32 for _ in range(MAX_TEXT_LENGTH-len(ord_text))]
    # converting ascii codes to their byte represantation
    ord_text = [("0"*BYTE+"{0:b}".format(num))[-BYTE:] for num in ord_text]
    # converting it to 1's and 0's
    return np.array(list(map(int, "".join(ord_text))))


def _encode_image_with_text(org_data: np.ndarray,
                            text: str,
                            key: int,
                            nnoise: int) -> Image.Image:
    """Encodes a text in an image.

    Parameters
    ----------
    org_data : numpy.ndarray
        The image that will keep the text in it.
    text : str
        The piece of text that will be encoded.
    key : int
        A key that will be used for RNG. It's will be needed 
        when extracting the text.
    nnoise : int
        For further protection from the people with the original image,
        "salt" is added. This is a coefficent to determine the size of the salt.
        salt size = nnoise * size of the encoded data.

    Returns
    -------
    PIL.Image.Image
        An image that has text encoded in it.
    """
    # the function works with the flattened data
    data = org_data.reshape(np.prod(org_data.shape)).copy()
    # the key is used for generating the random indexes in the flattened data
    np.random.seed(key)
    if MAX_TEXT_LENGTH*BYTE > data.shape[0]:
        raise Exception("Error: Image size is to small for the text.")
    idxs = np.random.choice(data.shape[0], MAX_TEXT_LENGTH*BYTE, replace=False)
    # converting the text to 1's and 0's for storing
    hidden = _text_to_binary(text)
    # putting the data to the last bit
    data[idxs] = (data[idxs] >> 1 << 1) + hidden

    # if someone else has the original image, part of the text could be
    # uncovered, so now will add random noise(salt?) to some other indexes
    # nnoise is a variable that controls the number of noise is added
    # more nnoise is more secure but also means more changes to the original
    if nnoise*MAX_TEXT_LENGTH*BYTE > data.shape[0]:
        nnoise = data.shape[0]//(MAX_TEXT_LENGTH*BYTE)
        warnings.warn("Warning: Lowering the nnoise.")

    noise_idxs = np.random.choice(
        data.shape[0], nnoise*MAX_TEXT_LENGTH*BYTE, replace=False)
    # filtering the ones thats in the hidden text's indexes
    noise_idxs = np.setdiff1d(noise_idxs, idxs)
    # adding the noise if any space available
    if len(noise_idxs) > 0:
        n = len(noise_idxs)
        data[noise_idxs[:n]] = (data[noise_idxs[:n]] >> 1 << 1) + 1
        data[noise_idxs[n:]] = data[noise_idxs[n:]] >> 1 << 1
    # return the reshaped flattened data with the original shape
    return Image.fromarray(data.reshape(org_data.shape))


def _decode_image(encoded_data: np.ndarray, key: int) -> str:
    """Extracts the text hidden in an image.

    Parameters
    ----------
    encoded_data : numpy.ndarray
        The image that has the text in it.
    key : int
        A key that is used for RNG. It's should be the
        same with the one used when encoding the text.

    Returns
    -------
    str
        The extracted text.
    """
    # flattenes the encoded image
    data = encoded_data.reshape(np.prod(encoded_data.shape)).copy()
    # sets the key and gets the indexes of the hidden text
    np.random.seed(key)
    idxs = np.random.choice(data.shape[0], MAX_TEXT_LENGTH*BYTE, replace=False)
    # takes the leftmost bits from the indexes
    s = "".join(map(str, data[idxs] % 2))
    # splits the bits into groups of bytes
    s = [s[i:i+BYTE] for i in range(0, len(s), BYTE)]
    # first, converts each of the bytes to ascii code then to chars
    # and gets the encoded text within the image
    s = "".join([chr(int(bin_, base=2)) for bin_ in s])
    # removes paddings and returns the text
    return s.strip()
